@@ -44,6 +44,7 @@ from nbxmpp.modules.iq import BaseIq
 from nbxmpp.modules.nickname import Nickname
 from nbxmpp.modules.delay import Delay
 from nbxmpp.modules.muc import MUC
+from nbxmpp.modules.muc.moderation import Moderation
 from nbxmpp.modules.idle import Idle
 from nbxmpp.modules.pgplegacy import PGPLegacy
 from nbxmpp.modules.vcard_avatar import VCardAvatar
@@ -56,7 +57,9 @@ from nbxmpp.modules.tune import Tune
 from nbxmpp.modules.mood import Mood
 from nbxmpp.modules.location import Location
 from nbxmpp.modules.user_avatar import UserAvatar
-from nbxmpp.modules.bookmarks import Bookmarks
+from nbxmpp.modules.bookmarks.private_bookmarks import PrivateBookmarks
+from nbxmpp.modules.bookmarks.pep_bookmarks import PEPBookmarks
+from nbxmpp.modules.bookmarks.native_bookmarks import NativeBookmarks
 from nbxmpp.modules.openpgp import OpenPGP
 from nbxmpp.modules.omemo import OMEMO
 from nbxmpp.modules.annotations import Annotations
@@ -75,9 +78,15 @@ from nbxmpp.modules.chatstates import Chatstates
 from nbxmpp.modules.register import Register
 from nbxmpp.modules.http_upload import HTTPUpload
 from nbxmpp.modules.mam import MAM
+from nbxmpp.modules.vcard_temp import VCardTemp
+from nbxmpp.modules.vcard4 import VCard4
+from nbxmpp.modules.ping import Ping
+from nbxmpp.modules.delimiter import Delimiter
+from nbxmpp.modules.roster import Roster
+from nbxmpp.modules.last_activity import LastActivity
+from nbxmpp.modules.entity_time import EntityTime
 from nbxmpp.modules.misc import unwrap_carbon
 from nbxmpp.modules.misc import unwrap_mam
-from nbxmpp.structs import StanzaTimeoutError
 from nbxmpp.util import get_properties_struct
 from nbxmpp.util import get_invalid_xml_regex
 from nbxmpp.util import is_websocket_close
@@ -148,6 +157,7 @@ class StanzaDispatcher(Observable):
         self._modules['HTTPAuth'] = HTTPAuth(self._client)
         self._modules['Nickname'] = Nickname(self._client)
         self._modules['MUC'] = MUC(self._client)
+        self._modules['Moderation'] = Moderation(self._client)
         self._modules['Delay'] = Delay(self._client)
         self._modules['Captcha'] = Captcha(self._client)
         self._modules['Idle'] = Idle(self._client)
@@ -161,7 +171,9 @@ class StanzaDispatcher(Observable):
         self._modules['Tune'] = Tune(self._client)
         self._modules['Location'] = Location(self._client)
         self._modules['UserAvatar'] = UserAvatar(self._client)
-        self._modules['Bookmarks'] = Bookmarks(self._client)
+        self._modules['PrivateBookmarks'] = PrivateBookmarks(self._client)
+        self._modules['PEPBookmarks'] = PEPBookmarks(self._client)
+        self._modules['NativeBookmarks'] = NativeBookmarks(self._client)
         self._modules['OpenPGP'] = OpenPGP(self._client)
         self._modules['OMEMO'] = OMEMO(self._client)
         self._modules['Annotations'] = Annotations(self._client)
@@ -180,6 +192,13 @@ class StanzaDispatcher(Observable):
         self._modules['Register'] = Register(self._client)
         self._modules['HTTPUpload'] = HTTPUpload(self._client)
         self._modules['MAM'] = MAM(self._client)
+        self._modules['VCardTemp'] = VCardTemp(self._client)
+        self._modules['VCard4'] = VCard4(self._client)
+        self._modules['Ping'] = Ping(self._client)
+        self._modules['Delimiter'] = Delimiter(self._client)
+        self._modules['Roster'] = Roster(self._client)
+        self._modules['LastActivity'] = LastActivity(self._client)
+        self._modules['EntityTime'] = EntityTime(self._client)
 
         for instance in self._modules.values():
             for handler in instance.handlers:
@@ -224,7 +243,7 @@ class StanzaDispatcher(Observable):
             self._parser.Parse(data)
         except (ExpatError, ValueError) as error:
             self._log.error('XML parsing error: %s', error)
-            self.notify('parsing-error', error)
+            self.notify('parsing-error', str(error))
             return
 
         # end stream:stream tag received
@@ -358,11 +377,11 @@ class StanzaDispatcher(Observable):
             return
 
         own_jid = self._client.get_bound_jid()
-        properties = get_properties_struct(name)
+        properties = get_properties_struct(name, own_jid)
 
         if name == 'iq':
             if stanza.getFrom() is None and own_jid is not None:
-                stanza.setFrom(own_jid.getBare())
+                stanza.setFrom(own_jid.bare)
 
         if name == 'message':
             # https://tools.ietf.org/html/rfc6120#section-8.1.1.1
@@ -373,13 +392,14 @@ class StanzaDispatcher(Observable):
             to = stanza.getTo()
             if to is None:
                 stanza.setTo(own_jid)
-            elif not to.bareMatch(own_jid):
+
+            elif not to.bare_match(own_jid):
                 self._log.warning('Message addressed to someone else: %s',
                                   stanza)
                 return
 
             if stanza.getFrom() is None:
-                stanza.setFrom(own_jid.getBare())
+                stanza.setFrom(own_jid.bare)
 
             # Unwrap carbon
             try:
@@ -483,7 +503,7 @@ class StanzaDispatcher(Observable):
 
             if timeout < time.monotonic():
                 self._id_callbacks.pop(id_)
-                func(self._client, StanzaTimeoutError(id_), **user_data)
+                func(self._client, None, **user_data)
         return True
 
     def _remove_timeout_source(self):
@@ -491,7 +511,7 @@ class StanzaDispatcher(Observable):
             GLib.source_remove(self._timeout_id)
             self._timeout_id = None
 
-    def remove_ping_callback(self, id_):
+    def remove_iq_callback(self, id_):
         self._id_callbacks.pop(id_, None)
 
     def clear_iq_callbacks(self):
