@@ -18,11 +18,22 @@ data structures, including jabber-objects like JID or different stanzas and
 sub- stanzas) handling routines
 """
 
+from __future__ import annotations
+
+from typing import Any
+from typing import Union
+from typing import Optional
+from typing import cast
+
 import time
 import hashlib
-import socket
 import functools
+import warnings
 from base64 import b64encode
+from dataclasses import dataclass
+from dataclasses import asdict
+
+from gi.repository import GLib
 
 import idna
 from precis_i18n import get_profile
@@ -32,191 +43,14 @@ from nbxmpp.namespaces import Namespace
 def ascii_upper(s):
     return s.upper()
 
-NS_ACTIVITY       = 'http://jabber.org/protocol/activity'             # XEP-0108
-NS_ADDRESS        = 'http://jabber.org/protocol/address'              # XEP-0033
-NS_AGENTS         = 'jabber:iq:agents'
-NS_AMP            = 'http://jabber.org/protocol/amp'
-NS_AMP_ERRORS     = NS_AMP + '#errors'
-NS_ARCHIVE        = 'urn:xmpp:archive'                                # XEP-0136
-NS_ARCHIVE_AUTO   = NS_ARCHIVE + ':auto'                              # XEP-0136
-NS_ARCHIVE_MANAGE = NS_ARCHIVE + ':manage'                            # XEP-0136
-NS_ARCHIVE_MANUAL = NS_ARCHIVE + ':manual'                            # XEP-0136
-NS_ARCHIVE_PREF   = NS_ARCHIVE + ':pref'
-NS_ATOM           = 'http://www.w3.org/2005/Atom'
-NS_ATTENTION      = 'urn:xmpp:attention:0'                            # XEP-0224
-NS_AUTH           = 'jabber:iq:auth'
-NS_AVATAR         = 'http://www.xmpp.org/extensions/xep-0084.html#ns-metadata'
-NS_AVATAR_METADATA = 'urn:xmpp:avatar:metadata'                       # XEP-0084
-NS_AVATAR_DATA    = 'urn:xmpp:avatar:data'                            # XEP-0084
-NS_BIND           = 'urn:ietf:params:xml:ns:xmpp-bind'
-NS_BLOCKING       = 'urn:xmpp:blocking'                               # XEP-0191
-NS_BOB            = 'urn:xmpp:bob'                                    # XEP-0231
-NS_BOOKMARKS      = 'storage:bookmarks'                               # XEP-0048
-NS_BOOKMARKS_2    = 'urn:xmpp:bookmarks:0'                            # XEP-0402
-NS_BOOKMARKS_COMPAT = 'urn:xmpp:bookmarks:0#compat'                   # XEP-0402
-NS_BROWSE         = 'jabber:iq:browse'
-NS_BROWSING       = 'http://jabber.org/protocol/browsing'             # XEP-0195
-NS_BYTESTREAM     = 'http://jabber.org/protocol/bytestreams'          # XEP-0065
-NS_CAPS           = 'http://jabber.org/protocol/caps'                 # XEP-0115
-NS_CAPTCHA        = 'urn:xmpp:captcha'                                # XEP-0158
-NS_CARBONS        = 'urn:xmpp:carbons:2'                              # XEP-0280
-NS_CHATMARKERS    = 'urn:xmpp:chat-markers:0'                         # XEP-0333
-NS_CHATSTATES     = 'http://jabber.org/protocol/chatstates'           # XEP-0085
-NS_CHATTING       = 'http://jabber.org/protocol/chatting'             # XEP-0194
-NS_CLIENT         = 'jabber:client'
-NS_CONDITIONS     = 'urn:xmpp:muc:conditions:1'                       # XEP-0306
-NS_COMMANDS       = 'http://jabber.org/protocol/commands'
-NS_COMPONENT_ACCEPT = 'jabber:component:accept'
-NS_COMPONENT_1    = 'http://jabberd.jabberstudio.org/ns/component/1.0'
-NS_COMPRESS       = 'http://jabber.org/protocol/compress'             # XEP-0138
-NS_CONFERENCE     = 'jabber:x:conference'
-NS_CORRECT        = 'urn:xmpp:message-correct:0'                      # XEP-0308
-NS_DATA           = 'jabber:x:data'                                   # XEP-0004
-NS_DATA_MEDIA     = 'urn:xmpp:media-element'                          # XEP-0221
-NS_DELAY          = 'jabber:x:delay'
-NS_DELAY2         = 'urn:xmpp:delay'
-NS_DIALBACK       = 'jabber:server:dialback'
-NS_DISCO          = 'http://jabber.org/protocol/disco'
-NS_DISCO_INFO     = NS_DISCO + '#info'
-NS_DISCO_ITEMS    = NS_DISCO + '#items'
-NS_EME            = 'urn:xmpp:eme:0'                                  # XEP-0380
-NS_ENCRYPTED      = 'jabber:x:encrypted'                              # XEP-0027
-NS_ESESSION       = 'http://www.xmpp.org/extensions/xep-0116.html#ns'
-NS_ESESSION_INIT  = 'http://www.xmpp.org/extensions/xep-0116.html#ns-init' # XEP-0116
-NS_EVENT          = 'jabber:x:event'                                  # XEP-0022
-NS_FEATURE        = 'http://jabber.org/protocol/feature-neg'
-NS_FILE           = 'http://jabber.org/protocol/si/profile/file-transfer' # XEP-0096
-NS_FORWARD        = 'urn:xmpp:forward:0'                              # XEP-0297
-NS_GAMING         = 'http://jabber.org/protocol/gaming'               # XEP-0196
-NS_GATEWAY        = 'jabber:iq:gateway'                               # XEP-0100
-NS_GEOLOC         = 'http://jabber.org/protocol/geoloc'               # XEP-0080
-NS_GROUPCHAT      = 'gc-1.0'
-NS_MSG_HINTS      = 'urn:xmpp:hints'                                  # XEP-0280
-NS_HTTP_AUTH      = 'http://jabber.org/protocol/http-auth'            # XEP-0070
-NS_HTTP_BIND      = 'http://jabber.org/protocol/httpbind'             # XEP-0124
-NS_HTTPUPLOAD     = 'urn:xmpp:http:upload'                            # XEP-0363
-NS_HTTPUPLOAD_0   = 'urn:xmpp:http:upload:0'                          # XEP-0363
-NS_IBB            = 'http://jabber.org/protocol/ibb'
-NS_IDLE           = 'urn:xmpp:idle:1'                                 # XEP-0319
-NS_INVISIBLE      = 'presence-invisible'                              # Jabberd2
-NS_IQ             = 'iq'                                              # Jabberd2
-NS_JINGLE         ='urn:xmpp:jingle:1'                                # XEP-0166
-NS_JINGLE_ERRORS  = 'urn:xmpp:jingle:errors:1'                        # XEP-0166
-NS_JINGLE_RTP     = 'urn:xmpp:jingle:apps:rtp:1'                      # XEP-0167
-NS_JINGLE_RTP_AUDIO = 'urn:xmpp:jingle:apps:rtp:audio'                # XEP-0167
-NS_JINGLE_RTP_VIDEO = 'urn:xmpp:jingle:apps:rtp:video'                # XEP-0167
-NS_JINGLE_FILE_TRANSFER = 'urn:xmpp:jingle:apps:file-transfer:3'      # XEP-0234
-NS_JINGLE_FILE_TRANSFER_5 = 'urn:xmpp:jingle:apps:file-transfer:5'    # XEP-0234
-NS_JINGLE_XTLS='urn:xmpp:jingle:security:xtls:0'                      # XTLS: EXPERIMENTAL security layer of jingle
-NS_JINGLE_RAW_UDP = 'urn:xmpp:jingle:transports:raw-udp:1'            # XEP-0177
-NS_JINGLE_ICE_UDP = 'urn:xmpp:jingle:transports:ice-udp:1'            # XEP-0176
-NS_JINGLE_BYTESTREAM ='urn:xmpp:jingle:transports:s5b:1'              # XEP-0260
-NS_JINGLE_IBB     = 'urn:xmpp:jingle:transports:ibb:1'                # XEP-0261
-NS_LAST           = 'jabber:iq:last'
-NS_LOCATION       = 'http://jabber.org/protocol/geoloc'               # XEP-0080
-NS_MAM            = 'urn:xmpp:mam:0'                                  # XEP-0313
-NS_MAM_1          = 'urn:xmpp:mam:1'                                  # XEP-0313
-NS_MAM_2          = 'urn:xmpp:mam:2'                                  # XEP-0313
-NS_MESSAGE        = 'message'                                         # Jabberd2
-NS_MOOD           = 'http://jabber.org/protocol/mood'                 # XEP-0107
-NS_MUC            = 'http://jabber.org/protocol/muc'
-NS_MUC_USER       = NS_MUC + '#user'
-NS_MUC_ADMIN      = NS_MUC + '#admin'
-NS_MUC_OWNER      = NS_MUC + '#owner'
-NS_MUC_UNIQUE     = NS_MUC + '#unique'
-NS_MUC_CONFIG     = NS_MUC + '#roomconfig'
-NS_MUC_REQUEST    = NS_MUC + '#request'
-NS_MUC_INFO       = NS_MUC + '#roominfo'
-NS_NICK           = 'http://jabber.org/protocol/nick'                 # XEP-0172
-NS_OFFLINE        = 'http://www.jabber.org/jeps/jep-0030.html'        # XEP-0013
-NS_PHYSLOC        = 'http://jabber.org/protocol/physloc'              # XEP-0112
-NS_PING           = 'urn:xmpp:ping'                                   # XEP-0199
-NS_PRESENCE       = 'presence'                                        # Jabberd2
-NS_PRIVACY        = 'jabber:iq:privacy'
-NS_PRIVATE        = 'jabber:iq:private'
-NS_PROFILE        = 'http://jabber.org/protocol/profile'              # XEP-0154
-NS_PUBSUB         = 'http://jabber.org/protocol/pubsub'               # XEP-0060
-NS_PUBSUB_EVENT   = 'http://jabber.org/protocol/pubsub#event'
-NS_PUBSUB_PUBLISH_OPTIONS = NS_PUBSUB + '#publish-options'            # XEP-0060
-NS_PUBSUB_OWNER   = 'http://jabber.org/protocol/pubsub#owner'         # XEP-0060
-NS_PUBSUB_CONFIG  = 'http://jabber.org/protocol/pubsub#node_config'   # XEP-0060
-NS_REGISTER       = 'jabber:iq:register'
-NS_REGISTER_FEATURE = 'http://jabber.org/features/iq-register'
-NS_ROSTER         = 'jabber:iq:roster'
-NS_ROSTERNOTES    = 'storage:rosternotes'
-NS_ROSTERX        = 'http://jabber.org/protocol/rosterx'              # XEP-0144
-NS_ROSTER_VER     = 'urn:xmpp:features:rosterver'                     # XEP-0273
-NS_RPC            = 'jabber:iq:rpc'                                   # XEP-0009
-NS_RSM            = 'http://jabber.org/protocol/rsm'
-NS_SASL           = 'urn:ietf:params:xml:ns:xmpp-sasl'
-NS_SECLABEL       = 'urn:xmpp:sec-label:0'
-NS_SECLABEL_CATALOG = 'urn:xmpp:sec-label:catalog:2'
-NS_SEARCH         = 'jabber:iq:search'
-NS_SERVER         = 'jabber:server'
-NS_SESSION        = 'urn:ietf:params:xml:ns:xmpp-session'
-NS_SI             = 'http://jabber.org/protocol/si'                   # XEP-0096
-NS_SI_PUB         = 'http://jabber.org/protocol/sipub'                # XEP-0137
-NS_SID            = 'urn:xmpp:sid:0'                                  # XEP-0359
-NS_SIGNED         = 'jabber:x:signed'                                 # XEP-0027
-NS_SIMS           = 'urn:xmpp:sims:1'                                 # XEP-0385
-NS_SSN            = 'urn:xmpp:ssn'                                    # XEP-0155
-NS_STANZA_CRYPTO  = 'http://www.xmpp.org/extensions/xep-0200.html#ns' # XEP-0200
-NS_STANZAS        = 'urn:ietf:params:xml:ns:xmpp-stanzas'
-NS_STREAM         = 'http://affinix.com/jabber/stream'
-NS_STREAMS        = 'http://etherx.jabber.org/streams'
-NS_TIME           = 'jabber:iq:time'                                  # XEP-0900
-NS_TIME_REVISED   = 'urn:xmpp:time'                                   # XEP-0202
-NS_TLS            = 'urn:ietf:params:xml:ns:xmpp-tls'
-NS_TUNE           = 'http://jabber.org/protocol/tune'                 # XEP-0118
-NS_VACATION       = 'http://jabber.org/protocol/vacation'
-NS_VCARD          = 'vcard-temp'
-NS_GMAILNOTIFY    = 'google:mail:notify'
-NS_GTALKSETTING   = 'google:setting'
-NS_VCARD_UPDATE   = NS_VCARD + ':x:update'
-NS_VERSION        = 'jabber:iq:version'
-NS_VIEWING        = 'http://jabber.org/protocol/viewing'              # XEP--197
-NS_WAITINGLIST    = 'http://jabber.org/protocol/waitinglist'          # XEP-0130
-NS_XHTML_IM       = 'http://jabber.org/protocol/xhtml-im'             # XEP-0071
-NS_XHTML          = 'http://www.w3.org/1999/xhtml'                    # "
-NS_X_OOB          = 'jabber:x:oob'                                    # XEP-0066
-NS_DATA_LAYOUT    = 'http://jabber.org/protocol/xdata-layout'         # XEP-0141
-NS_DATA_VALIDATE  = 'http://jabber.org/protocol/xdata-validate'       # XEP-0122
-NS_XMPP_STREAMS   = 'urn:ietf:params:xml:ns:xmpp-streams'
-NS_RECEIPTS       = 'urn:xmpp:receipts'
-NS_PUBKEY_PUBKEY  = 'urn:xmpp:pubkey:2'                                              # XEP-0189
-NS_PUBKEY_REVOKE  = 'urn:xmpp:revoke:2'
-NS_PUBKEY_ATTEST  = 'urn:xmpp:attest:2'
-NS_STREAM_MGMT    = 'urn:xmpp:sm:3'                                   # XEP-198
-NS_HASHES         = 'urn:xmpp:hashes:1'                               # XEP-300
-NS_HASHES_2       = 'urn:xmpp:hashes:2'                               # XEP-300
-NS_HASHES_MD5     = 'urn:xmpp:hash-function-text-names:md5'
-NS_HASHES_SHA1    = 'urn:xmpp:hash-function-text-names:sha-1'
-NS_HASHES_SHA256  = 'urn:xmpp:hash-function-text-names:sha-256'
-NS_HASHES_SHA512  = 'urn:xmpp:hash-function-text-names:sha-512'
-NS_HASHES_SHA3_256 = 'urn:xmpp:hash-function-text-names:sha3-256'
-NS_HASHES_SHA3_512 = 'urn:xmpp:hash-function-text-names:sha3-512'
-NS_HASHES_BLAKE2B_256 = 'urn:xmpp:hash-function-text-names:id-blake2b256'
-NS_HASHES_BLAKE2B_512 = 'urn:xmpp:hash-function-text-names:id-blake2b512'
-NS_OPENPGP = 'urn:xmpp:openpgp:0'
-NS_OPENPGP_PK = 'urn:xmpp:openpgp:0:public-keys'
-NS_OPENPGP_SK = 'urn:xmpp:openpgp:0:secret-key'
-NS_OMEMO_TEMP = 'eu.siacs.conversations.axolotl'
-NS_OMEMO_TEMP_DL = 'eu.siacs.conversations.axolotl.devicelist'
-NS_OMEMO_TEMP_BUNDLE = 'eu.siacs.conversations.axolotl.bundles'
-NS_BOOKMARK_CONVERSION = 'urn:xmpp:bookmarks-conversion:0'
-NS_DOMAIN_BASED_NAME = 'urn:xmpp:domain-based-name:1'
-NS_HINTS = 'urn:xmpp:hints'
-NS_MUCLUMBUS = 'https://xmlns.zombofant.net/muclumbus/search/1.0'
-NS_FRAMING = 'urn:ietf:params:xml:ns:xmpp-framing'
-
-
 SASL_AUTH_MECHS = [
+    'SCRAM-SHA-512',
     'SCRAM-SHA-256-PLUS',
     'SCRAM-SHA-256',
     'SCRAM-SHA-1-PLUS',
     'SCRAM-SHA-1',
-    'PLAIN',
     'GSSAPI',
+    'PLAIN',
     'EXTERNAL',
     'ANONYMOUS',
 ]
@@ -415,29 +249,9 @@ _errorcodes = {
     'cancel': 'invalid-from'
 }
 
-_status_conditions = {
-    'realjid-public': 100,
-    'affiliation-changed': 101,
-    'unavailable-shown': 102,
-    'unavailable-not-shown': 103,
-    'configuration-changed': 104,
-    'self-presence': 110,
-    'logging-enabled': 170,
-    'logging-disabled': 171,
-    'non-anonymous': 172,
-    'semi-anonymous': 173,
-    'fully-anonymous': 174,
-    'room-created': 201,
-    'nick-assigned': 210,
-    'banned': 301,
-    'new-nick': 303,
-    'kicked': 307,
-    'removed-affiliation': 321,
-    'removed-membership': 322,
-    'removed-shutdown': 332,
-}
 
 _localpart_disallowed_chars = set('"&\'/:<>@')
+_localpart_escape_chars = ' "&\'/:<>@'
 
 
 STREAM_NOT_AUTHORIZED = 'urn:ietf:params:xml:ns:xmpp-streams not-authorized'
@@ -658,34 +472,12 @@ stream_exceptions = {'bad-format': BadFormat,
                      'xml-not-well-formed': XMLNotWellFormed}
 
 
+def deprecation_warning(message):
+    warnings.warn(message, DeprecationWarning)
+
+
 @functools.lru_cache(maxsize=None)
-def parse_jid(jid):
-    # https://tools.ietf.org/html/rfc7622#section-3.2
-
-    # Remove any portion from the first '/' character to the end of the
-    # string (if there is a '/' character present).
-
-    # Remove any portion from the beginning of the string to the first
-    # '@' character (if there is an '@' character present).
-
-    if jid.find('/') != -1:
-        rest, resourcepart = jid.split('/', 1)
-        resourcepart = validate_resourcepart(resourcepart)
-    else:
-        rest, resourcepart = jid, ''
-
-    if rest.find('@') != -1:
-        localpart, domainpart = rest.split('@', 1)
-        localpart = validate_localpart(localpart)
-    else:
-        localpart, domainpart = '', rest
-
-    domainpart = validate_domainpart(domainpart)
-    return localpart, domainpart, resourcepart
-
-
-@functools.lru_cache(maxsize=32768)
-def validate_localpart(localpart):
+def validate_localpart(localpart: str) -> str:
     if not localpart or len(localpart.encode()) > 1023:
         raise LocalpartByteLimit
 
@@ -699,8 +491,8 @@ def validate_localpart(localpart):
         raise LocalpartNotAllowedChar
 
 
-@functools.lru_cache(maxsize=32768)
-def validate_resourcepart(resourcepart):
+@functools.lru_cache(maxsize=None)
+def validate_resourcepart(resourcepart: str) -> str:
     if not resourcepart or len(resourcepart.encode()) > 1023:
         raise ResourcepartByteLimit
 
@@ -711,202 +503,253 @@ def validate_resourcepart(resourcepart):
         raise ResourcepartNotAllowedChar
 
 
-@functools.lru_cache(maxsize=4096)
-def validate_domainpart(domainpart):
+@functools.lru_cache(maxsize=None)
+def validate_domainpart(domainpart: str) -> str:
     if not domainpart:
         raise DomainpartByteLimit
 
-    # Check if this is a IPV4 address
-    try:
-        socket.inet_aton(domainpart)
-        return domainpart
-    except Exception:
-        pass
+    ip_address = domainpart.strip('[]')
+    if GLib.hostname_is_ip_address(ip_address):
+        return ip_address
 
-    # Check if this is a IPV6 address
-    if domainpart.startswith('[') and domainpart.endswith(']'):
-        try:
-            socket.inet_pton(socket.AF_INET6, domainpart.strip('[]'))
-            return domainpart
-        except Exception:
-            pass
+    length = len(domainpart.encode())
+    if length == 0 or length > 1023:
+        raise DomainpartByteLimit
 
     if domainpart.endswith('.'):  # RFC7622, 3.2
         domainpart = domainpart[:-1]
 
     try:
-        domainpart = idna.encode(domainpart, uts46=True)
+        idna_encode(domainpart)
     except Exception:
         raise DomainpartNotAllowedChar
 
-    length = len(domainpart)
-    if length == 0 or length > 1023:
-        raise DomainpartByteLimit
-
-    return domainpart.decode()
+    return domainpart
 
 
+@functools.lru_cache(maxsize=None)
+def idna_encode(domain: str) -> str:
+    return idna.encode(domain, uts46=True).decode()
+
+
+@functools.lru_cache(maxsize=None)
+def escape_localpart(localpart: str) -> str:
+    # https://xmpp.org/extensions/xep-0106.html#bizrules-algorithm
+    #
+    # If there are any instances of character sequences that correspond
+    # to escapings of the disallowed characters
+    # (e.g., the character sequence "\27") or the escaping character
+    # (i.e., the character sequence "\5c") in the source address,
+    # the leading backslash character MUST be escaped to the character
+    # sequence "\5c"
+
+    for char in '\\' + _localpart_escape_chars:
+        seq = "\\{:02x}".format(ord(char))
+        localpart = localpart.replace(seq, "\\5c{:02x}".format(ord(char)))
+
+    # Escape all other chars
+    for char in _localpart_escape_chars:
+        localpart = localpart.replace(char, "\\{:02x}".format(ord(char)))
+
+    return localpart
+
+
+@functools.lru_cache(maxsize=None)
+def unescape_localpart(localpart: str) -> str:
+    if localpart.startswith('\\20') or localpart.endswith('\\20'):
+        # Escaped JIDs are not allowed to start or end with \20
+        # so this localpart must be already unescaped
+        return localpart
+
+    for char in _localpart_escape_chars:
+        seq = "\\{:02x}".format(ord(char))
+        localpart = localpart.replace(seq, char)
+
+    for char in _localpart_escape_chars + "\\":
+        seq = "\\5c{:02x}".format(ord(char))
+        localpart = localpart.replace(seq, "\\{:02x}".format(ord(char)))
+
+    return localpart
+
+
+@dataclass(frozen=True)
 class JID:
-    """
-    JID can be built from string, modified, compared, serialised into string
-    """
+    localpart: Optional[str] = None
+    domain: Optional[str] = None
+    resource: Optional[str] = None
 
-    def __init__(self, jid=None, node='', domain='', resource=''):
-        """
-        JID can be specified as string (jid argument) or as separate parts
+    def __init__(self,
+                 localpart: Optional[str] = None,
+                 domain: Optional[str] = None,
+                 resource: Optional[str] = None):
 
-        Examples:
-        JID('node@domain/resource')
-        JID(node='node',domain='domain.org')
-        """
-        if not jid and not domain:
-            raise ValueError('JID must contain at least domain name')
+        if localpart is not None:
+            localpart = validate_localpart(localpart)
+            object.__setattr__(self, "localpart", localpart)
 
-        if isinstance(jid, JID):
-            self.node, self.domain = jid.node, jid.domain
-            self.resource = jid.resource
-        elif domain:
-            self.node, self.domain, self.resource = node, domain, resource
+        domain = validate_domainpart(domain)
+        object.__setattr__(self, "domain", domain)
+
+        if resource is not None:
+            resource = validate_resourcepart(resource)
+            object.__setattr__(self, "resource", resource)
+
+    @classmethod
+    @functools.lru_cache(maxsize=None)
+    def from_string(cls, jid_string: str, force_bare: bool = False) -> JID:
+        # https://tools.ietf.org/html/rfc7622#section-3.2
+
+        # Remove any portion from the first '/' character to the end of the
+        # string (if there is a '/' character present).
+
+        # Remove any portion from the beginning of the string to the first
+        # '@' character (if there is an '@' character present).
+
+        if jid_string.find('/') != -1:
+            rest, resourcepart = jid_string.split('/', 1)
         else:
-            self.node, self.domain, self.resource = parse_jid(jid)
+            rest, resourcepart = jid_string, None
 
-    def copy(self):
-        return JID(self)
-
-    def getNode(self):
-        """
-        Return the node part of the JID
-        """
-        return self.node
-
-    def setNode(self, node):
-        """
-        Set the node part of the JID to new value. Specify None to remove
-        the node part
-        """
-        self.node = node.lower()
-
-    def getDomain(self):
-        """
-        Return the domain part of the JID
-        """
-        return self.domain
-
-    def setDomain(self, domain):
-        """
-        Set the domain part of the JID to new value
-        """
-        self.domain = domain.lower()
-
-    def getResource(self):
-        """
-        Return the resource part of the JID
-        """
-        return self.resource
-
-    def setResource(self, resource):
-        """
-        Set the resource part of the JID to new value.
-        Specify None to remove the resource part
-        """
-        self.resource = resource
-
-    def getStripped(self):
-        """
-        Return the bare representation of JID. I.e. string value w/o resource
-        """
-        if not self.node:
-            return self.domain
-        return '%s@%s' % (self.node, self.domain)
-
-    def getBare(self):
-        """
-        Return the bare representation of JID. I.e. string value w/o resource
-        """
-        if not self.node:
-            return self.domain
-        return '%s@%s' % (self.node, self.domain)
-
-    def setBare(self):
-        """
-        Deletes the resource which makes the JID bare
-        """
-        self.resource = ''
-
-    @property
-    def isBare(self):
-        return self.node and self.domain and not self.resource
-
-    @property
-    def isDomain(self):
-        return self.domain and not self.node and not self.resource
-
-    @property
-    def isFull(self):
-        return bool(self.domain and self.node and self.resource)
-
-    def __eq__(self, other):
-        """
-        Compare the JID to another instance or to string for equality
-        """
-        if not isinstance(other, JID):
-            try:
-                other = JID(other)
-            except Exception:
-                return False
-
-        return (self.node == other.node and
-                self.domain == other.domain and
-                self.resource == other.resource)
-
-    def __ne__(self, other):
-        """
-        Compare the JID to another instance or to string for non-equality
-        """
-        return not self.__eq__(other)
-
-    def bareMatch(self, other):
-        """
-        Compare the node and domain parts of the JID's for equality
-        """
-        if not isinstance(other, JID):
-            try:
-                other = JID(other)
-            except Exception:
-                return False
-
-        return (self.node == other.node and
-                self.domain == other.domain)
-
-    def __str__(self):
-        """
-        Serialise JID into string
-        """
-        if self.node:
-            jid = '%s@%s' % (self.node, self.domain)
+        if rest.find('@') != -1:
+            localpart, domainpart = rest.split('@', 1)
         else:
-            jid = self.domain
+            localpart, domainpart = None, rest
 
-        if self.resource:
-            return '%s/%s' % (jid, self.resource)
+        if force_bare:
+            if localpart is None:
+                raise LocalpartByteLimit
+            resourcepart = None
+
+        return cls(localpart=localpart,
+                   domain=domainpart,
+                   resource=resourcepart)
+
+    @classmethod
+    @functools.lru_cache(maxsize=None)
+    def from_user_input(cls, user_input: str, escaped: bool = False) -> JID:
+        # Use this if we want JIDs to be escaped according to XEP-0106
+        # The standard JID parsing cannot be applied because user_input is
+        # not a valid JID.
+
+        # Only user_input which after escaping result in a bare JID can be
+        # successfully parsed.
+
+        # The assumpution is user_input is a bare JID so we start with an
+        # rsplit on @ because we assume there is no resource, so the char @
+        # in the localpart can later be escaped.
+
+        if escaped:
+            # for convenience
+            return cls.from_string(user_input)
+
+        if '@' in user_input:
+            localpart, domainpart = user_input.rsplit('@', 1)
+            if localpart.startswith(' ') or localpart.endswith(' '):
+                raise LocalpartNotAllowedChar
+
+            localpart = escape_localpart(localpart)
+
+        else:
+            localpart = None
+            domainpart = user_input
+
+        return cls(localpart=localpart,
+                   domain=domainpart,
+                   resource=None)
+
+    def __str__(self) -> str:
+        if self.localpart:
+            jid = f'{self.localpart}@{self.domain}'
+        else:
+            jid = cast(str, self.domain)
+
+        if self.resource is not None:
+            return f'{jid}/{self.resource}'
         return jid
 
     def __hash__(self):
-        """
-        Produce hash of the JID, Allows to use JID objects as keys of the
-        dictionary
-        """
         return hash(str(self))
 
-    def __repr__(self):
-        return str(self)
+    def __eq__(self, other: Union[str, JID]) -> bool:
+        if isinstance(other, str):
+            try:
+                return JID.from_string(other) == self
+            except Exception:
+                return False
+
+        return (self.localpart == other.localpart and
+                self.domain == other.domain and
+                self.resource == other.resource)
+
+    def __ne__(self, other: Union[str, JID]) -> bool:
+        return not self.__eq__(other)
+
+    def domain_to_ascii(self) -> str:
+        return idna_encode(self.domain)
+
+    @property
+    def bare(self) -> str:
+        if self.localpart is not None:
+            return f'{self.localpart}@{self.domain}'
+        return self.domain
+
+    @property
+    def is_bare(self) -> bool:
+        return self.resource is None
+
+    def new_as_bare(self) -> JID:
+        if self.resource is None:
+            return self
+        new = asdict(self)
+        new.pop('resource')
+        return JID(**new)
+
+    def bare_match(self, other: Union[str, JID]) -> bool:
+        if isinstance(other, str):
+            other = JID.from_string(other)
+        return self.bare == other.bare
+
+    @property
+    def is_domain(self) -> bool:
+        return self.localpart is None and self.resource is None
+
+    @property
+    def is_full(self) -> bool:
+        return (self.localpart is not None and
+                self.domain is not None and
+                self.resource is not None)
+
+    def new_with(self, **kwargs: dict[str, str]) -> JID:
+        new = asdict(self)
+        new.update(kwargs)
+        return JID(**new)
+
+    def to_user_string(self, show_punycode: bool = True) -> str:
+        domain = self.domain_to_ascii()
+        if domain.startswith('xn--') and show_punycode:
+            domain_encoded = f' ({domain})'
+        else:
+            domain_encoded = ''
+
+        if self.localpart is None:
+            return f'{self}{domain_encoded}'
+
+        localpart = unescape_localpart(self.localpart)
+
+        if self.resource is None:
+            return f'{localpart}@{self.domain}{domain_encoded}'
+        return f'{localpart}@{self.domain}/{self.resource}{domain_encoded}'
+
+    def copy(self) -> JID:
+        deprecation_warning('copy() is not needed, JID is immutable')
+        return self
 
 
 class StreamErrorNode(Node):
     def __init__(self, node):
         Node.__init__(self, node=node)
 
-        self._text = {}
+        self._text: dict[Optional[str], str] = {}
 
         text_elements = self.getTags('text', namespace=Namespace.XMPP_STREAMS)
         for element in text_elements:
@@ -914,14 +757,14 @@ class StreamErrorNode(Node):
             text = element.getData()
             self._text[lang] = text
 
-    def get_condition(self):
+    def get_condition(self) -> Optional[str]:
         for tag in self.getChildren():
             if (tag.getName() != 'text' and
                     tag.getNamespace() == Namespace.XMPP_STREAMS):
                 return tag.getName()
         return None
 
-    def get_text(self, pref_lang=None):
+    def get_text(self, pref_lang: Optional[str] = None) -> str:
         if pref_lang is not None:
             text = self._text.get(pref_lang)
             if text is not None:
@@ -1003,6 +846,12 @@ class Protocol(Node):
         if timestamp is not None:
             self.setTimestamp(timestamp)
 
+    def isError(self):
+        return self.getAttr('type') == 'error'
+
+    def isResult(self):
+        return self.getAttr('type') == 'result'
+
     def getTo(self):
         """
         Return value of the 'to' attribute
@@ -1045,46 +894,50 @@ class Protocol(Node):
         """
         attr = self.getAttr('jid')
         if attr:
-            return JID(attr)
+            return JID.from_string(attr)
         return attr
 
-    def getID(self):
+    def getID(self) -> Optional[str]:
         """
         Return the value of the 'id' attribute
         """
         return self.getAttr('id')
 
-    def setTo(self, val):
+    def setTo(self, val: Union[str, JID]):
         """
         Set the value of the 'to' attribute
         """
-        self.setAttr('to', JID(val))
+        if not isinstance(val, JID):
+            val = JID.from_string(val)
+        self.setAttr('to', val)
 
-    def getType(self):
+    def getType(self) -> Optional[str]:
         """
         Return the value of the 'type' attribute
         """
         return self.getAttr('type')
 
-    def setFrom(self, val):
+    def setFrom(self, val: Union[str, JID]):
         """
         Set the value of the 'from' attribute
         """
-        self.setAttr('from', JID(val))
+        if not isinstance(val, JID):
+            val = JID.from_string(val)
+        self.setAttr('from', val)
 
-    def setType(self, val):
+    def setType(self, val: str):
         """
         Set the value of the 'type' attribute
         """
         self.setAttr('type', val)
 
-    def setID(self, val):
+    def setID(self, val: str):
         """
         Set the value of the 'id' attribute
         """
         self.setAttr('id', val)
 
-    def getError(self):
+    def getError(self) -> Optional[str]:
         """
         Return the error-condition (if present) or the textual description
         of the error (otherwise)
@@ -1195,7 +1048,11 @@ class Protocol(Node):
                 props.append(prop)
         return props
 
-    def getTag(self, name, attrs=None, namespace=None, protocol=False):
+    def getTag(self,
+               name: str,
+               attrs: Optional[dict[str, Any]] = None,
+               namespace: Optional[str] = None,
+               protocol: bool = False) -> Optional[Node]:
         """
         Return the Node instance for the tag.
         If protocol is True convert to a new Protocol/Message instance.
@@ -1207,13 +1064,14 @@ class Protocol(Node):
             return Protocol(node=tag)
         return tag
 
-    def __setitem__(self, item, val):
+    def __setitem__(self, item: str, val: Union[str, JID]):
         """
         Set the item 'item' to the value 'val'
         """
         if item in ['to', 'from']:
-            val = JID(val)
-        return self.setAttr(item, val)
+            if not isinstance(val, JID):
+                val = JID.from_string(val)
+        self.setAttr(item, val)
 
 
 class Message(Protocol):
@@ -1442,7 +1300,7 @@ class Presence(Protocol):
         """
         Return the status string of the message
         """
-        return self.getTagData('status')
+        return self.getTagData('status') or ''
 
     def setPriority(self, val):
         """
@@ -1534,14 +1392,14 @@ class Iq(Protocol):
     """
 
     def __init__(self,
-                 typ=None,
-                 queryNS=None,
-                 attrs=None,
-                 to=None,
-                 frm=None,
-                 payload=None,
-                 xmlns=Namespace.CLIENT,
-                 node=None):
+                 typ: Optional[str] = None,
+                 queryNS: Optional[str] = None,
+                 attrs: Optional[dict[str, str]] = None,
+                 to: Optional[Union[JID, str]] = None,
+                 frm: Optional[Union[JID, str]] = None,
+                 payload: Optional[list[Union[Node, str]]] = None,
+                 xmlns: str = Namespace.CLIENT,
+                 node: Optional[Node] = None):
         """
         You can specify type, query namespace any additional attributes,
         recipient of the iq, sender of the iq, any additional payload (f.e.
@@ -1563,7 +1421,7 @@ class Iq(Protocol):
         if queryNS:
             self.setQueryNS(queryNS)
 
-    def getQuery(self):
+    def getQuery(self) -> Optional[Node]:
         """
         Return the IQ's child element if it exists, None otherwise.
         """
@@ -1573,7 +1431,7 @@ class Iq(Protocol):
             return children[0]
         return None
 
-    def getQueryNS(self):
+    def getQueryNS(self) -> Optional[str]:
         """
         Return the namespace of the 'query' child element
         """
@@ -1582,7 +1440,7 @@ class Iq(Protocol):
             return tag.getNamespace()
         return None
 
-    def getQuerynode(self):
+    def getQuerynode(self) -> Optional[str]:
         """
         Return the 'node' attribute value of the 'query' child element
         """
@@ -1591,16 +1449,7 @@ class Iq(Protocol):
             return tag.getAttr('node')
         return None
 
-    def getQueryPayload(self):
-        """
-        Return the 'query' child element payload
-        """
-        tag = self.getQuery()
-        if tag:
-            return tag.getPayload()
-        return None
-
-    def getQueryChildren(self):
+    def getQueryChildren(self) -> Optional[list[Node]]:
         """
         Return the 'query' child element child nodes
         """
@@ -1609,7 +1458,7 @@ class Iq(Protocol):
             return tag.getChildren()
         return None
 
-    def getQueryChild(self, name=None):
+    def getQueryChild(self, name: Optional[str] = None) -> Optional[Node]:
         """
         Return the 'query' child element with name, or the first element
         which is not an error element
@@ -1626,7 +1475,7 @@ class Iq(Protocol):
                     return node
         return None
 
-    def setQuery(self, name=None):
+    def setQuery(self, name: Optional[str] = None) -> Node:
         """
         Change the name of the query node, creating it if needed. Keep the
         existing name if none is given (use 'query' if it's a creation).
@@ -1639,25 +1488,25 @@ class Iq(Protocol):
             query.setName(name)
         return query
 
-    def setQueryNS(self, namespace):
+    def setQueryNS(self, namespace: str) -> None:
         """
         Set the namespace of the 'query' child element
         """
         self.setQuery().setNamespace(namespace)
 
-    def setQueryPayload(self, payload):
+    def setQueryPayload(self, payload: list[Union[Node, str]]):
         """
         Set the 'query' child element payload
         """
         self.setQuery().setPayload(payload)
 
-    def setQuerynode(self, node):
+    def setQuerynode(self, node: str) -> None:
         """
         Set the 'node' attribute value of the 'query' child element
         """
         self.setQuery().setAttr('node', node)
 
-    def buildReply(self, typ):
+    def buildReply(self, typ: str) -> Iq:
         """
         Build and return another Iq object of specified type. The to, from and
         query child node of new Iq are pre-set as reply to this Iq.
@@ -1669,6 +1518,12 @@ class Iq(Protocol):
         iq.setQuery(self.getQuery().getName()).setNamespace(self.getQueryNS())
         return iq
 
+    def buildSimpleReply(self, typ: str) -> Iq:
+        return Iq(typ,
+                  to=self.getFrom(),
+                  attrs={'id': self.getID()})
+
+
 class Hashes(Node):
     """
     Hash elements for various XEPs as defined in XEP-300
@@ -1790,13 +1645,15 @@ class Hashes2(Node):
 
 
 class BindRequest(Iq):
-    def __init__(self, resource):
-        if resource is not None:
-            resource = Node('resource', payload=resource)
+    def __init__(self, resource: Optional[str]):
+        if resource is None:
+            res = resource
+        else:
+            res = Node('resource', payload=resource)
         Iq.__init__(self, typ='set')
         self.addChild(node=Node('bind',
                                 {'xmlns': Namespace.BIND},
-                                payload=resource))
+                                payload=res))
 
 
 class TLSRequest(Node):
@@ -1811,7 +1668,7 @@ class SessionRequest(Iq):
 
 
 class StreamHeader(Node):
-    def __init__(self, domain, lang=None):
+    def __init__(self, domain: str, lang: Optional[str] = None):
         if lang is None:
             lang = 'en'
         Node.__init__(self,
@@ -1824,7 +1681,7 @@ class StreamHeader(Node):
 
 
 class WebsocketOpenHeader(Node):
-    def __init__(self, domain, lang=None):
+    def __init__(self, domain: str, lang: Optional[str] = None):
         if lang is None:
             lang = 'en'
         Node.__init__(self,
@@ -1840,7 +1697,7 @@ class WebsocketCloseHeader(Node):
 
 
 class Features(Node):
-    def __init__(self, node):
+    def __init__(self, node: Node):
         Node.__init__(self, node=node)
 
     def has_starttls(self):
@@ -1853,7 +1710,7 @@ class Features(Node):
     def has_sasl(self):
         return self.getTag('mechanisms', namespace=Namespace.SASL) is not None
 
-    def get_mechs(self):
+    def get_mechs(self) -> set[str]:
         mechanisms = self.getTag('mechanisms', namespace=Namespace.SASL)
         if mechanisms is None:
             return set()
@@ -1939,24 +1796,18 @@ class Error(Protocol):
         If the 'node' is not the received stanza but locally created ('to' and
         'from' fields needs not swapping) specify the 'reply' argument as false.
         """
+        if node.getType() == 'error':
+            raise ValueError('Can’t create error reply from error')
+
         if reply:
             Protocol.__init__(self,
                               to=node.getFrom(),
-                              frm=node.getTo(),
                               node=node)
+            self.delAttr('from')
         else:
             Protocol.__init__(self, node=node)
         self.setError(error)
-        if node.getType() == 'error':
-            self.__str__ = self.__dupstr__
 
-    def __dupstr__(self, _dup1=None, _dup2=None):
-        """
-        Dummy function used as preventor of creating error node in reply to
-        error node. I.e. you will not be able to serialise "double" error
-        into string.
-        """
-        return ''
 
 class DataField(Node):
     """
